"""
    xkbregistry
    ~~~~~~~~~~~

    CFFI-based libxkbregistry bindings for Python.  See README.rst for details.

    :copyright: Copyright 2023 by Stephen Early <steve@assorted.org.uk>
    :license: MIT, see LICENSE for details.
"""



# NB these tests are intended to check whether the python bindings are
# working, not whether libxkbregistry itself is working!

# We assume that the xkb-data package is installed

from unittest import TestCase

from xkbregistry import rxkb

import os

# A directory that is guaranteed to exist
testdir = os.path.dirname(os.path.abspath(__file__))
# A path that is guaranteed not to exist
nonexistent = os.path.join(testdir, "must-not-exist")
# A file that is guaranteed to exist
testfile = os.path.abspath(__file__)


class TestContext(TestCase):
    def test_create(self):
        self.assertIsInstance(rxkb.Context(), rxkb.Context)
        self.assertIsInstance(rxkb.Context(no_default_includes=True),
                              rxkb.Context)
        self.assertIsInstance(rxkb.Context(load_exotic_rules=True),
                              rxkb.Context)

    def test_include_path_append(self):
        ctx = rxkb.Context(no_default_includes=True)
        self.assertIsNone(ctx.include_path_append(testdir))
        with self.assertRaises(rxkb.RXKBPathError):
            ctx.include_path_append(nonexistent)
        with self.assertRaises(rxkb.RXKBPathError):
            ctx.include_path_append(testfile)

    def test_include_path_append_default(self):
        ctx = rxkb.Context(no_default_includes=True)
        self.assertIsNone(ctx.include_path_append_default())
        # I can't figure out a way to get the library call to fail to
        # test the failure case!

    def test_parse(self):
        ctx = rxkb.Context()
        ctx.parse("base")
        with self.assertRaises(rxkb.RXKBAlreadyParsed):
            ctx.parse("base")

    def test_parse_default_ruleset(self):
        ctx = rxkb.Context()
        ctx.parse_default_ruleset()
        with self.assertRaises(rxkb.RXKBAlreadyParsed):
            ctx.parse_default_ruleset()

    def test_models(self):
        ctx = rxkb.Context()
        self.assertIn("pc102", ctx.models)
        self.assertEqual(ctx.models["pc102"].description,
                         "Generic 102-key PC")
        self.assertEqual(ctx.models["pc102"].vendor, "Generic")
        self.assertEqual(ctx.models["pc102"].popularity,
                         rxkb.Popularity.RXKB_POPULARITY_STANDARD)

    def test_layouts(self):
        ctx = rxkb.Context()
        self.assertIn("us", ctx.layouts)
        x = ctx.layouts["us"]
        self.assertIsNone(x.variant)
        self.assertEqual(x.fullname, "us")
        self.assertEqual(x.brief, "en")
        self.assertEqual(x.description, "English (US)")
        self.assertSetEqual(x.iso639_codes, {"eng"})
        self.assertEqual(x.popularity,
                         rxkb.Popularity.RXKB_POPULARITY_STANDARD)
        self.assertIn("nec_vndr/jp", ctx.layouts)
        self.assertEqual(ctx.layouts["nec_vndr/jp"].iso3166_codes, {"JP"})

    def test_layout_variant(self):
        ctx = rxkb.Context()
        self.assertIn("us(chr)", ctx.layouts)
        x = ctx.layouts["us(chr)"]
        self.assertEqual(x.fullname, "us(chr)")
        self.assertEqual(x.variant, "chr")
        self.assertEqual(x.brief, "chr")
        self.assertEqual(x.description, "Cherokee")
        self.assertSetEqual(x.iso639_codes, {"chr"})

    def test_groups(self):
        ctx = rxkb.Context()
        # Groups are not required to have names. We are looking for
        # the group called "lv2", because it is short!
        for group in ctx.option_groups:
            if group.name == "lv2":
                self.assertEqual(
                    group.description, "Key to choose the 2nd level")
                self.assertEqual(group.popularity,
                                 rxkb.Popularity.RXKB_POPULARITY_STANDARD)
                self.assertIn("lv2:lsgt_switch", group.options)
                x = group.options["lv2:lsgt_switch"]
                self.assertEqual(x.description, 'The "< >" key')
                self.assertEqual(x.popularity,
                                 rxkb.Popularity.RXKB_POPULARITY_STANDARD)
                break
        else:
            self.fail(msg="lv2 option group not found")



from cffi import FFI
ffibuilder = FFI()

# Currently implemented with reference to libxkbcommon-1.0

ffibuilder.set_source("xkbregistry._ffi", """
#include <stdarg.h>
#include <xkbcommon/xkbregistry.h>

static void _log_handler(void *user_data,
                         enum rxkb_log_level level,
                         const char *message);

static void _log_handler_internal(
    struct rxkb_context *context,
    enum rxkb_log_level level,
    const char *format,
    va_list args)
{
  char buf[1024];
  void *user_data;

  user_data=rxkb_context_get_user_data(context);
  vsnprintf(buf, 1024, format, args);
  buf[1023]=0;
  _log_handler(user_data, level, buf);
}

void _set_log_handler_internal(struct rxkb_context *context)
{
  rxkb_context_set_log_fn(context, _log_handler_internal);
}

""",
                      libraries=['xkbregistry'])

ffibuilder.cdef("""
typedef ... FILE;
typedef ... va_list;

struct rxkb_context;
struct rxkb_model;
struct rxkb_layout;
struct rxkb_option_group;
struct rxkb_option;
struct rxkb_iso639_code;
struct rxkb_iso3166_code;

enum rxkb_popularity {
    RXKB_POPULARITY_STANDARD = ...,
    RXKB_POPULARITY_EXOTIC = ...
};

enum rxkb_context_flags {
  RXKB_CONTEXT_NO_FLAGS = ...,
  RXKB_CONTEXT_NO_DEFAULT_INCLUDES = ...,
  RXKB_CONTEXT_LOAD_EXOTIC_RULES = ...,
};

enum rxkb_log_level {
  RXKB_LOG_LEVEL_CRITICAL = ...,
  RXKB_LOG_LEVEL_ERROR = ...,
  RXKB_LOG_LEVEL_WARNING = ...,
  RXKB_LOG_LEVEL_INFO = ...,
  RXKB_LOG_LEVEL_DEBUG = ...
};

struct rxkb_context *
rxkb_context_new(enum rxkb_context_flags flags);

void
rxkb_context_set_log_level(struct rxkb_context *ctx,
                           enum rxkb_log_level level);

enum rxkb_log_level
rxkb_context_get_log_level(struct rxkb_context *ctx);

void
rxkb_context_set_log_fn(struct rxkb_context *ctx,
                         void(*log_fn)(struct rxkb_context *ctx,
                                       enum rxkb_log_level level,
                                       const char *format, va_list args));

bool
rxkb_context_parse(struct rxkb_context *ctx, const char *ruleset);

bool
rxkb_context_parse_default_ruleset(struct rxkb_context *ctx);

struct rxkb_context *
rxkb_context_ref(struct rxkb_context *ctx);

struct rxkb_context *
rxkb_context_unref(struct rxkb_context *ctx);

void
rxkb_context_set_user_data(struct rxkb_context *ctx, void *user_data);

void *
rxkb_context_get_user_data(struct rxkb_context *ctx);

bool
rxkb_context_include_path_append(struct rxkb_context *ctx, const char *path);

bool
rxkb_context_include_path_append_default(struct rxkb_context *ctx);

struct rxkb_model *
rxkb_model_first(struct rxkb_context *ctx);

struct rxkb_model *
rxkb_model_next(struct rxkb_model *m);

struct rxkb_model *
rxkb_model_ref(struct rxkb_model *m);

struct rxkb_model *
rxkb_model_unref(struct rxkb_model *m);

const char *
rxkb_model_get_name(struct rxkb_model *m);

const char *
rxkb_model_get_description(struct rxkb_model *m);

const char *
rxkb_model_get_vendor(struct rxkb_model *m);

enum rxkb_popularity
rxkb_model_get_popularity(struct rxkb_model *m);

struct rxkb_layout *
rxkb_layout_first(struct rxkb_context *ctx);

struct rxkb_layout *
rxkb_layout_next(struct rxkb_layout *l);

struct rxkb_layout *
rxkb_layout_ref(struct rxkb_layout *l);

struct rxkb_layout *
rxkb_layout_unref(struct rxkb_layout *l);

const char *
rxkb_layout_get_name(struct rxkb_layout *l);

const char *
rxkb_layout_get_variant(struct rxkb_layout *l);

const char *
rxkb_layout_get_brief(struct rxkb_layout *l);

const char *
rxkb_layout_get_description(struct rxkb_layout *l);

enum rxkb_popularity
rxkb_layout_get_popularity(struct rxkb_layout *l);

struct rxkb_option_group *
rxkb_option_group_first(struct rxkb_context *ctx);

struct rxkb_option_group *
rxkb_option_group_next(struct rxkb_option_group *g);

struct rxkb_option_group *
rxkb_option_group_ref(struct rxkb_option_group *g);

struct rxkb_option_group *
rxkb_option_group_unref(struct rxkb_option_group *g);

const char *
rxkb_option_group_get_name(struct rxkb_option_group *m);

const char *
rxkb_option_group_get_description(struct rxkb_option_group *m);

bool
rxkb_option_group_allows_multiple(struct rxkb_option_group *g);

enum rxkb_popularity
rxkb_option_group_get_popularity(struct rxkb_option_group *g);

struct rxkb_option *
rxkb_option_first(struct rxkb_option_group *group);

struct rxkb_option *
rxkb_option_next(struct rxkb_option *o);

struct rxkb_option *
rxkb_option_ref(struct rxkb_option *o);

struct rxkb_option *
rxkb_option_unref(struct rxkb_option *o);

const char *
rxkb_option_get_name(struct rxkb_option *o);

const char *
rxkb_option_get_brief(struct rxkb_option *o);

const char *
rxkb_option_get_description(struct rxkb_option *o);

enum rxkb_popularity
rxkb_option_get_popularity(struct rxkb_option *o);

struct rxkb_iso639_code *
rxkb_iso639_code_ref(struct rxkb_iso639_code *iso639);

struct rxkb_iso639_code *
rxkb_iso639_code_unref(struct rxkb_iso639_code *iso639);

const char *
rxkb_iso639_code_get_code(struct rxkb_iso639_code *iso639);

struct rxkb_iso639_code *
rxkb_layout_get_iso639_first(struct rxkb_layout *layout);

struct rxkb_iso639_code *
rxkb_iso639_code_next(struct rxkb_iso639_code *iso639);

struct rxkb_iso3166_code *
rxkb_iso3166_code_ref(struct rxkb_iso3166_code *iso3166);

struct rxkb_iso3166_code *
rxkb_iso3166_code_unref(struct rxkb_iso3166_code *iso3166);

const char *
rxkb_iso3166_code_get_code(struct rxkb_iso3166_code *iso3166);

struct rxkb_iso3166_code *
rxkb_layout_get_iso3166_first(struct rxkb_layout *layout);

struct rxkb_iso3166_code *
rxkb_iso3166_code_next(struct rxkb_iso3166_code *iso3166);

void _set_log_handler_internal(struct rxkb_context *context);

extern "Python" void _log_handler(void *user_data,
                                  enum rxkb_log_level level,
                                  const char *message);

void free(void *ptr);

""")

if __name__ == "__main__":
    ffibuilder.compile(verbose=True)



import enum

from xkbregistry._ffi import ffi, lib


class _keepref:
    """Function wrapper that keeps a reference to another object."""
    def __init__(self, ref, func):
        self.ref = ref
        self.func = func

    def __call__(self, *args, **kwargs):
        self.func(*args, **kwargs)


def _string_or_none(r):
    if r != ffi.NULL:
        return ffi.string(r).decode('utf8')


class RXKBError(Exception):
    """Base for all RXKB exceptions"""
    pass


class RXKBAlreadyParsed(RXKBError):
    """The context has already parsed a ruleset."""
    pass


class RXKBPathError(RXKBError):
    """There was a problem altering the include path of a Context."""
    pass


class RXKBParseError(RXKBError):
    """The ruleset could not be parsed."""
    pass


# Internal helper for logging callback
def _onerror_do_nothing(exception, exc_value, traceback):
    return


@ffi.def_extern(onerror=_onerror_do_nothing)
def _log_handler(user_data, level, message):
    context = ffi.from_handle(user_data)
    if context._log_fn:
        context._log_fn(context, level, ffi.string(message).decode('utf8'))


@enum.unique
class Popularity(enum.IntEnum):
    """Describes the popularity of an item.

    Historically, some highly specialized or experimental definitions
    are excluded from the default list and shipped in separate
    files. If these extra definitions are loaded (see
    Context(load_exotic_rules=True)), the popularity of the item is
    set accordingly.

    If the exotic items are not loaded, all items will have the
    standard popularity.
    """
    RXKB_POPULARITY_STANDARD = lib.RXKB_POPULARITY_STANDARD
    RXKB_POPULATITY_EXOTIC = lib.RXKB_POPULARITY_EXOTIC


class Context:
    """xkbregistry library context

    The context contains general library state, like include paths and
    parsed data. Objects are created in a specific context, and
    multiple contexts may coexist simultaneously. Objects from
    different contexts are completely separated and do not share any
    memory or state.
    """
    def __init__(self, no_default_includes=False, load_exotic_rules=False):
        flags = lib.RXKB_CONTEXT_NO_FLAGS
        if no_default_includes:
            flags = flags | lib.RXKB_CONTEXT_NO_DEFAULT_INCLUDES
        if load_exotic_rules:
            flags = flags | lib.RXKB_CONTEXT_LOAD_EXOTIC_RULES
        context = lib.rxkb_context_new(flags)
        if not context:
            raise RXKBError("Couldn't create RXKB context")
        self._context = ffi.gc(context, _keepref(lib, lib.rxkb_context_unref))
        self._log_fn = None
        # We keep a reference to the handle to keep it alive
        self._userdata = ffi.new_handle(self)
        lib.rxkb_context_set_user_data(self._context, self._userdata)
        self._parsed = False

    def include_path_append(self, path):
        "Append a new entry to the context's include path."
        if self._parsed:
            raise RXKBAlreadyParsed()
        r = lib.rxkb_context_include_path_append(
            self._context, path.encode('utf8'))
        if r != 1:
            raise RXKBPathError("Failed to append to include path")

    def include_path_append_default(self):
        "Append the default include paths to the context's include path."
        if self._parsed:
            raise RXKBAlreadyParsed()
        r = lib.rxkb_context_include_path_append_default(self._context)
        if r != 1:
            raise RXKBPathError("Failed to append default include paths")

    def parse(self, ruleset):
        "Parse the given ruleset"
        if self._parsed:
            raise RXKBAlreadyParsed()
        r = lib.rxkb_context_parse(self._context, ruleset.encode('utf8'))
        if r != 1:
            raise RXKBParseError()
        self._parsed = True

    def parse_default_ruleset(self):
        "Parse the default ruleset as configured at build time"
        if self._parsed:
            raise RXKBAlreadyParsed()
        r = lib.rxkb_context_parse_default_ruleset(self._context)
        if r != 1:
            raise RXKBParseError()
        self._parsed = True

    @property
    def models(self):
        "Dictionary mapping model name to Model object"
        if not hasattr(self, '_models'):
            self._models = {}
            if not self._parsed:
                self.parse_default_ruleset()
            model = lib.rxkb_model_first(self._context)
            while model != ffi.NULL:
                x = Model(model)
                self._models[x.name] = x
                model = lib.rxkb_model_next(model)
        return self._models

    @property
    def layouts(self):
        "Dictionary mapping layout fullname to Layout object"
        if not hasattr(self, '_layouts'):
            self._layouts = {}
            if not self._parsed:
                self.parse_default_ruleset()
            layout = lib.rxkb_layout_first(self._context)
            while layout != ffi.NULL:
                x = Layout(layout)
                self._layouts[x.fullname] = x
                layout = lib.rxkb_layout_next(layout)
        return self._layouts

    @property
    def option_groups(self):
        "List of OptionGroup objects"
        if not hasattr(self, '_option_groups'):
            self._option_groups = []
            if not self._parsed:
                self.parse_default_ruleset()
            option_group = lib.rxkb_option_group_first(self._context)
            while option_group != ffi.NULL:
                self._option_groups.append(OptionGroup(option_group))
                option_group = lib.rxkb_option_group_next(option_group)
        return self._option_groups


class Model:
    """An XKB model
    """
    def __init__(self, model):
        self.name = ffi.string(
            lib.rxkb_model_get_name(model)).decode('ascii')
        self.description = _string_or_none(
            lib.rxkb_model_get_description(model))
        self.vendor = _string_or_none(
            lib.rxkb_model_get_vendor(model))
        self.popularity = Popularity(lib.rxkb_model_get_popularity(model))

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"rxkb.Model('{self.name}')"


class Layout:
    """An XKB layout, including an optional variant

    Where the variant is None, the layout is the base layout.

    For example, "us" is the base layout, "us(intl)" is the "intl"
    variant of the layout "us".
    """
    def __init__(self, layout):
        self.name = ffi.string(
            lib.rxkb_layout_get_name(layout)).decode('ascii')
        self.variant = _string_or_none(
            lib.rxkb_layout_get_variant(layout))
        self.brief = _string_or_none(
            lib.rxkb_layout_get_brief(layout))
        self.description = _string_or_none(
            lib.rxkb_layout_get_description(layout))
        self.popularity = Popularity(lib.rxkb_layout_get_popularity(layout))
        self.iso639_codes = frozenset(self._codes(
            layout,
            lib.rxkb_layout_get_iso639_first,
            lib.rxkb_iso639_code_next,
            lib.rxkb_iso639_code_get_code))
        self.iso3166_codes = frozenset(self._codes(
            layout,
            lib.rxkb_layout_get_iso3166_first,
            lib.rxkb_iso3166_code_next,
            lib.rxkb_iso3166_code_get_code))

    @staticmethod
    def _codes(layout, first_fn, next_fn, get_fn):
        code = first_fn(layout)
        while code != ffi.NULL:
            yield ffi.string(get_fn(code)).decode('ascii')
            code = next_fn(code)

    @property
    def fullname(self):
        return f"{self.name}({self.variant})" if self.variant else self.name

    def __str__(self):
        return self.fullname

    def __repr__(self):
        return f"rxkb.Layout('{self.fullname}')"


class OptionGroup:
    """An XKB option group

    Option groups divide the individual options into logical
    groups. Their main purpose is to indicate whether some options are
    mutually exclusive or not.

    Option groups may have a name, but are not required to.
    """
    def __init__(self, option_group):
        self.name = _string_or_none(
            lib.rxkb_option_group_get_name(option_group))
        self.description = _string_or_none(
            lib.rxkb_option_group_get_description(option_group))
        self.allows_multiple = bool(
            lib.rxkb_option_group_allows_multiple(option_group))
        self.popularity = Popularity(
            lib.rxkb_option_group_get_popularity(option_group))
        self.options = {}
        option = lib.rxkb_option_first(option_group)
        while option != ffi.NULL:
            x = Option(option)
            self.options[x.name] = x
            option = lib.rxkb_option_next(option)

    def __repr__(self):
        if self.name:
            return f"rxkb.OptionGroup('{self.name}')"
        return super().__repr__()


class Option:
    """An XKB option
    """
    def __init__(self, option):
        self.name = ffi.string(
            lib.rxkb_option_get_name(option)).decode('ascii')
        self.brief = _string_or_none(
            lib.rxkb_option_get_brief(option))
        self.description = _string_or_none(
            lib.rxkb_option_get_description(option))
        self.popularity = Popularity(lib.rxkb_option_get_popularity(option))

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"rxkb.Option('{self.name}')"
